import pickle
from dataclasses import dataclass, field
from typing import Dict, Any, List

from intervaltree import IntervalTree, Interval

from .components import ExclusionInterval, ExclusionPoint, convert_min_bounds, convert_max_bounds


def get_mass_interval(ex_interval: ExclusionInterval):
    return Interval(convert_min_bounds(ex_interval.min_mass),
                    convert_max_bounds(ex_interval.max_mass),
                    ex_interval)


@dataclass
class MassIntervalTree:
    """
    ExclusionList store excluded intervals. Excluded intervals or stored in a 1D IntervalTree based on mass.
    Therefore, only a single interval can be stored for each unique mass interval, adding another interval
    with the same underlying mass interval will override the stored object
    """

    interval_tree: IntervalTree = field(default_factory=lambda: IntervalTree())
    id_dict: Dict[str, set] = field(default_factory=lambda: dict())

    def add(self, ex_interval: ExclusionInterval):
        if ex_interval.interval_id is None:
            raise Exception('Cannot add an interval with id = None')
        mass_interval = get_mass_interval(ex_interval)
        self.interval_tree.add(mass_interval)
        self.id_dict.setdefault(ex_interval.interval_id, set()).add(mass_interval)

    def remove(self, ex_interval: ExclusionInterval) -> List[ExclusionInterval]:
        mass_intervals = self._get_interval(ex_interval)
        if mass_intervals is None:
            return []

        for mass_interval in set(mass_intervals):
            self.interval_tree.remove(mass_interval)

        for mass_interval in mass_intervals:
            self.id_dict[mass_interval.data.interval_id].remove(mass_interval)

        return [mass_interval.data for mass_interval in mass_intervals]

    def _get_interval(self, ex_interval: ExclusionInterval) -> List[Interval]:
        if ex_interval.interval_id is None:
            # retrieve intervals by bounds
            mass_intervals = self._get_intervals_by_bounds(ex_interval)
        else:
            # retrieve intervals by id
            mass_intervals = self._get_intervals_by_id(ex_interval.interval_id)
            mass_intervals = [i for i in mass_intervals if i.data.is_enveloped_by(ex_interval)]

        return mass_intervals

    def _get_intervals_by_bounds(self, ex_interval: ExclusionInterval) -> List[Interval]:
        intervals = self.interval_tree.envelop(get_mass_interval(ex_interval))
        intervals = [i for i in intervals if i.data.is_enveloped_by(ex_interval)]
        return intervals

    def _get_intervals_by_id(self, interval_id: Any) -> List[Interval]:
        if interval_id in self.id_dict:
            return list(self.id_dict[interval_id])
        return []

    def is_excluded(self, point: ExclusionPoint) -> bool:

        for point in self.query_by_point(point):
            return True
        return False

    def query_by_interval(self, ex_interval: ExclusionInterval) -> List[ExclusionInterval]:
        return [mass_interval.data for mass_interval in self._get_interval(ex_interval)]

    def query_by_point(self, point: ExclusionPoint) -> List[ExclusionInterval]:
        if point.mass is None:
            intervals = [interval for interval in self.interval_tree]
        else:
            intervals = self.interval_tree[point.mass]

        return (interval.data for interval in intervals if point.is_bounded_by(interval.data))

    def query_by_id(self, interval_id: Any) -> List[ExclusionInterval]:
        return [interval.data for interval in self._get_intervals_by_id(interval_id)]

    def save(self, file_path: str):
        """
        Save a the interval tree as a pickled obj
        """
        with open(file_path, "wb") as file:
            pickle.dump(self.interval_tree, file, -1)

    def load(self, file_path: str) -> None:
        """
        Loads a pickled interval tree and then populates id_dict
        """
        with open(file_path, "rb") as file:
            self.interval_tree = pickle.load(file)
            self.id_dict = {}
            for interval in self.interval_tree:
                data = interval.data
                self.id_dict.setdefault(data.interval_id, set()).add(interval)

    def clear(self) -> None:
        """
        Clears all data
        """
        self.interval_tree.clear()
        self.id_dict = {}

    def __len__(self):
        return len(self.interval_tree)

    def stats(self):
        return {'len': len(self),
                'id_table_len': sum([len(self.id_dict[key]) for key in self.id_dict]),
                'class': str(type(self))}



import ast
import sys
from typing import Union, Dict, Tuple
from pydantic import BaseModel


def convert_min_bounds(min_bound: Union[float, None]) -> float:
    if min_bound is None:
        return -sys.float_info.max
    return min_bound


def convert_max_bounds(max_bound: Union[float, None]) -> float:
    if max_bound is None:
        return sys.float_info.max
    return max_bound


class ExclusionInterval(BaseModel):
    """
    Represents an interval in the excluded space.

    id: The id of the interval. Does not have to be unique. If None: Represents all IDs.
    charge: The charge of the excluded interval. If None: the Interval represents all charges
    min_bounds: The lower 'inclusive' bound of the interval. If None: Will be set to sys.float_info.min
    max_bounds: The upper 'exclusive' bound of the interval. If None: Will be set to sys.float_info.max
    """
    interval_id: Union[str, None]
    charge: Union[int, None]
    min_mass: Union[float, None]
    max_mass: Union[float, None]
    min_rt: Union[float, None]
    max_rt: Union[float, None]
    min_ook0: Union[float, None]
    max_ook0: Union[float, None]
    min_intensity: Union[float, None]
    max_intensity: Union[float, None]

    def is_enveloped_by(self, other: 'ExclusionInterval') -> bool:

        if other.charge is not None and self.charge is not None and self.charge != other.charge:
            return False

        if convert_min_bounds(self.min_mass) < convert_min_bounds(other.min_mass) or \
                convert_max_bounds(self.max_mass) > convert_max_bounds(other.max_mass):
            return False

        if convert_min_bounds(self.min_rt) < convert_min_bounds(other.min_rt) or \
                convert_max_bounds(self.max_rt) > convert_max_bounds(other.max_rt):
            return False

        if convert_min_bounds(self.min_ook0) < convert_min_bounds(other.min_ook0) or \
                convert_max_bounds(self.max_ook0) > convert_max_bounds(other.max_ook0):
            return False

        if convert_min_bounds(self.min_intensity) < convert_min_bounds(other.min_intensity) or \
                convert_max_bounds(self.max_intensity) > convert_max_bounds(other.max_intensity):
            return False

        return True

    def is_valid(self) -> bool:
        if convert_min_bounds(self.min_mass) > convert_max_bounds(self.max_mass):
            return False
        if convert_min_bounds(self.min_rt) > convert_max_bounds(self.max_rt):
            return False
        if convert_min_bounds(self.min_ook0) > convert_max_bounds(self.max_ook0):
            return False
        if convert_min_bounds(self.min_intensity) > convert_max_bounds(self.max_intensity):
            return False
        return True

    @staticmethod
    def from_str(serialized_interval: str) -> 'ExclusionInterval':
        res = ast.literal_eval(serialized_interval)
        return ExclusionInterval.from_dict(res)

    @staticmethod
    def from_dict(res: Dict) -> 'ExclusionInterval':
        exclusion_interval = ExclusionInterval(**res)
        return exclusion_interval


class ExclusionPoint(BaseModel):
    """
    Represents a point in the excluded space. None values will be ignored.
    """
    charge: Union[int, None]
    mass: Union[float, None]
    rt: Union[float, None]
    ook0: Union[float, None]
    intensity: Union[float, None]

    def is_bounded_by(self, interval: ExclusionInterval) -> bool:
        """
        Check if point given by is_excluded() is within interval
        """
        if self.charge is not None and interval.charge is not None and self.charge != interval.charge:
            return False

        if self.rt is not None and (self.rt < convert_min_bounds(interval.min_rt) or
                                    self.rt >= convert_max_bounds(interval.max_rt)):
            return False

        if self.ook0 is not None and (self.ook0 < convert_min_bounds(interval.min_ook0) or
                                      self.ook0 >= convert_max_bounds(interval.max_ook0)):
            return False

        if self.intensity is not None and (self.intensity < convert_min_bounds(interval.min_intensity) or
                                           self.intensity >= convert_max_bounds(interval.max_intensity)):
            return False

        if self.mass is not None and (self.mass < convert_min_bounds(interval.min_mass) or
                                      self.mass >= convert_max_bounds(interval.max_mass)):
            return False

        return True

    @staticmethod
    def from_str(serialized_point: str) -> 'ExclusionPoint':
        res = ast.literal_eval(serialized_point)
        return ExclusionPoint.from_dict(res)

    @staticmethod
    def from_dict(res: Dict) -> 'ExclusionPoint':
        return ExclusionPoint(**res)


class DynamicExclusionTolerance(BaseModel):
    charge: bool
    mass: Union[float, None]
    rt: Union[float, None]
    ook0: Union[float, None]
    intensity: Union[float, None]

    @staticmethod
    def from_str(serialized_tolerance: str) -> 'DynamicExclusionTolerance':
        res = ast.literal_eval(serialized_tolerance)
        return DynamicExclusionTolerance.from_dict(res)

    @staticmethod
    def from_dict(res: dict) -> 'DynamicExclusionTolerance':
        return DynamicExclusionTolerance(**res)

    def calculate_mass_bounds(self, mass: Union[float, None]) -> Tuple:
        if self.mass and mass:
            min_mass = mass - mass * self.mass / 1_000_000
            max_mass = mass + mass * self.mass / 1_000_000
            return min_mass, max_mass
        return None, None

    def calculate_rt_bounds(self, rt: Union[float, None]) -> Tuple:
        if self.rt and rt:
            min_rt = rt - self.rt
            max_rt = rt + self.rt
            return min_rt, max_rt
        return None, None

    def calculate_ook0_bounds(self, ook0: Union[float, None]) -> Tuple:
        if self.ook0 and ook0:
            min_ook0 = ook0 - self.ook0
            max_ook0 = ook0 + self.ook0
            return min_ook0, max_ook0
        return None, None

    def calculate_intensity_bounds(self, intensity: Union[float, None]) -> Tuple:
        if self.intensity and intensity:
            min_intensity = intensity - intensity * self.intensity
            max_intensity = intensity + intensity * self.intensity
            return min_intensity, max_intensity
        return None, None

    def construct_interval(self, interval_id: str, exclusion_point: ExclusionPoint) -> ExclusionInterval:

        charge = exclusion_point.charge
        if self.charge is False:
            charge = None

        min_mass, max_mass = self.calculate_mass_bounds(exclusion_point.mass)
        min_rt, max_rt = self.calculate_rt_bounds(exclusion_point.rt)
        min_ook0, max_ook0 = self.calculate_ook0_bounds(exclusion_point.ook0)
        min_intensity, max_intensity = self.calculate_intensity_bounds(exclusion_point.intensity)

        return ExclusionInterval(interval_id=interval_id, charge=charge, min_mass=min_mass,
                                 max_mass=max_mass, min_rt=min_rt, max_rt=max_rt, min_ook0=min_ook0,
                                 max_ook0=max_ook0, min_intensity=min_intensity, max_intensity=max_intensity)



import json
from typing import List, Dict

import requests

from .components import ExclusionPoint, ExclusionInterval


def clear(exclusion_api_ip: str, timeout=None) -> int:
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/clear',
                             timeout=timeout)
    response.raise_for_status()
    return json.loads(response.content)


def load(exclusion_api_ip: str, exid: str, timeout=None):
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/load?exid={exid}',
                             timeout=timeout)
    response.raise_for_status()


def save(exclusion_api_ip: str, exid: str, timeout=None):
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/save?exid={exid}',
                             timeout=timeout)
    response.raise_for_status()


def delete(exclusion_api_ip: str, exid: str, timeout=None):
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/delete?exid={exid}',
                             timeout=timeout)
    response.raise_for_status()


def get_statistics(exclusion_api_ip: str, timeout=None) -> Dict:
    response = requests.get(url=f'{exclusion_api_ip}/exclusionms/statistics',
                            timeout=timeout)
    response.raise_for_status()

    return json.loads(response.content)


def get_len(exclusion_api_ip: str, timeout=None) -> Dict:
    return get_statistics(exclusion_api_ip)['len']


def get_files(exclusion_api_ip: str, timeout=None) -> List[str]:
    response = requests.get(url=f'{exclusion_api_ip}/exclusionms/file',
                            timeout=timeout)
    response.raise_for_status()
    return json.loads(response.content)


def add_interval(exclusion_api_ip: str, exclusion_interval: ExclusionInterval, timeout=None) -> None:
    add_intervals(exclusion_api_ip, [exclusion_interval], timeout)


def add_intervals(exclusion_api_ip: str, exclusion_intervals: List[ExclusionInterval], timeout=None) -> None:
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/intervals',
                             json=[interval.dict() for interval in exclusion_intervals],
                             timeout=timeout)
    response.raise_for_status()


def search_interval(exclusion_api_ip: str, exclusion_interval: ExclusionInterval, timeout=None) -> List[
    ExclusionInterval]:
    return search_intervals(exclusion_api_ip, [exclusion_interval], timeout)[0]


def search_intervals(exclusion_api_ip: str, exclusion_intervals: List[ExclusionInterval], timeout=None) -> List[
    List[ExclusionInterval]]:
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/intervals/search',
                             json=[interval.dict() for interval in exclusion_intervals],
                             timeout=timeout)

    response.raise_for_status()

    response_dicts = json.loads(response.content)
    exclusion_intervals = [[ExclusionInterval.from_dict(d) for d in l] for l in response_dicts]
    return exclusion_intervals


def delete_interval(exclusion_api_ip: str, exclusion_interval: ExclusionInterval, timeout=None) -> List[
    ExclusionInterval]:
    return delete_intervals(exclusion_api_ip, [exclusion_interval], timeout)[0]


def delete_intervals(exclusion_api_ip: str, exclusion_intervals: List[ExclusionInterval], timeout=None) -> \
        List[List[ExclusionInterval]]:
    response = requests.delete(url=f'{exclusion_api_ip}/exclusionms/intervals',
                               json=[interval.dict() for interval in exclusion_intervals],
                               timeout=timeout)

    response.raise_for_status()

    response_dicts = json.loads(response.content)
    exclusion_intervals = [[ExclusionInterval.from_dict(d) for d in l] for l in response_dicts]
    return exclusion_intervals


def search_point(exclusion_api_ip: str, exclusion_point: ExclusionPoint, timeout=None) -> List[
    ExclusionInterval]:
    return search_points(exclusion_api_ip, [exclusion_point], timeout)[0]


def search_points(exclusion_api_ip: str, exclusion_points: List[ExclusionPoint], timeout=None) -> List[List[
    ExclusionInterval]]:
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/points/search',
                             json=[point.dict() for point in exclusion_points],
                             timeout=timeout)

    response.raise_for_status()

    response_dicts = json.loads(response.content)
    exclusion_intervals = [[ExclusionInterval.from_dict(d) for d in l] for l in response_dicts]
    return exclusion_intervals


def exclusion_search_point(exclusion_api_ip: str, exclusion_point: ExclusionPoint, timeout=None) -> bool:
    return exclusion_search_points(exclusion_api_ip, [exclusion_point], timeout)[0]


def exclusion_search_points(exclusion_api_ip: str, exclusion_points: List[ExclusionPoint], timeout=None) -> List[bool]:
    response = requests.post(url=f'{exclusion_api_ip}/exclusionms/points/exclusion_search',
                             json=[point.dict() for point in exclusion_points],
                             timeout=timeout)

    response.raise_for_status()

    return json.loads(response.content)

def is_connected(exclusion_api_ip: str) -> bool:
    try:
        get_statistics(exclusion_api_ip)
    except requests.exceptions.ConnectionError:
        return False

    return True



from pathlib import Path

from setuptools import setup

with open("README", "r", encoding="utf-8") as fh:
    long_description = fh.read()

req = 'requirements.txt'
if Path(req).is_file():
    with open(req) as f:
        requirements = f.read().splitlines()
else:
    requirements = []

setup(
    name='exclusionms',
    version='0.1.1',
    packages=['exclusionms'],
    url='',
    license='',
    author='Patrick Garrett',
    author_email='pgarrett@scripps.edu',
    description=long_description,
    install_requires=requirements,
    python_requires='>=3.6'
)
